import sys

from .validator import validate_first, validate_last, validate_ipv4, validate_ipv6, validate_timestamp
from .exceptions import NonUniqueCmdlineArgumentError, ArgumentError
from .constants import ip_argument_invalids


class Arguments:

    def __init__(self, terminal_arguments, mmap_possible):
        self.first = None
        self.last = None
        self.timestamps = None
        self.ipv4 = None
        self.ipv6 = None
        self.termargs = terminal_arguments
        self.termargs_length = len(terminal_arguments)
        self.mmap = mmap_possible
        self.skip = False

        self.enable_mmap()
        self.run()

    def unique(self, key):
        if self.termargs.count(key) - 1:
            NonUniqueCmdlineArgumentError(key)

    def enable_mmap(self):
        if self.mmap:
            # log file at end so no need to parse this
            self.termargs_length -= 1
            self.termargs = self.termargs[:self.termargs_length]

    def run(self):
        for index, argument in enumerate(self.termargs):
            if self.skip:
                self.skip = False
                continue

            if argument == '--first' or argument == '-f':
                self.unique(argument)
                if index + 1 >= self.termargs_length or not validate_first(self.termargs[index + 1]):
                    ArgumentError("Input a positive number for the --first command\n")

                self.first = int(self.termargs[index + 1])
                self.skip = True

            elif argument == '--last' or argument == '-l':
                self.unique(argument)
                if index + 1 >= self.termargs_length or not validate_last(self.termargs[index + 1]):
                    ArgumentError("Input a positive number for the --last command\n")

                self.last = int(self.termargs[index + 1])
                self.skip = True

            elif argument == '--timestamps' or argument == '-t':
                self.unique(argument)
                self.timestamps = True

            elif argument == '--ipv4' or argument == '-i':
                self.unique(argument)
                if index + 1 < self.termargs_length and (self.termargs[index + 1]not in ip_argument_invalids):
                    if not validate_ipv4(self.termargs[index + 1]):
                        ArgumentError("Input a valid ipv4 address for the --ipv4 command\n")

                    self.ipv4 = self.termargs[index + 1]
                    self.skip = True

                else:
                    self.ipv4 = True

            elif argument == '--ipv6' or argument == '-I':
                self.unique(argument)
                if index + 1 < self.termargs_length and (self.termargs[index + 1] not in ip_argument_invalids):
                    if not validate_ipv6(self.termargs[index + 1]):
                        ArgumentError("Input a valid ipv6 address for the --ipv6 command\n")

                    self.ipv6 = self.termargs[index + 1]
                    self.skip = True
                else:
                    self.ipv6 = True

            else:
                ArgumentError("You inputted an unknown commands\n")



#!/usr/bin/python3
import sys

from .parser import Parser
from .arguments import Arguments
from .exceptions import HelpError, LogReadError

"""
Redparser - parsing log file based upon command line argument

This main function is the entry point to the redparser library

"""
def main():

    # exclude the python interpreter
    termargs = sys.argv[1:]
    
    # length of terminal arguments to be used for redparser
    termargs_length = len(termargs)

    # check if help was called or no arguments supplied
    if '-h' in termargs or '--help' in termargs or termargs_length == 0:
        HelpError()

    try:
        with open(sys.argv[-1], 'rb') as file_obj:
            mmap_possible = 1
            args = Arguments(termargs, mmap_possible)
            p = Parser(file_obj, args)
            p.run()
    except Exception:
        try:
            if not sys.stdin.isatty():
                file_obj = sys.stdin.buffer.read()
                mmap_possible = 0
                args = Arguments(termargs, mmap_possible)
                p = Parser(file_obj, args)
                p.run()

        except Exception:
            LogReadError("Error reading log file please pipe data to python file or have log file at end of command\n")

if __name__ == '__main__':

    main()


import sys

from .constants import help

# define redparser user-defined exceptions
class RedParserError(Exception):
    """Base class for other exceptions"""
    pass

class HelpError():
    """
    Exception raised for command line with no arguments or with -h or --help in arguments
    in the command line arguments.

    Attributes:
        help -- redparser help output
    """
    def __init__(self, help=help):
        self.help = help
        print(self.help)
        # stop the program
        sys.exit(1)

class NonUniqueCmdlineArgumentError(HelpError):
    """
    Exception raised for errors in the command line arguments.

    Attributes:
        nonunique -- the command argument that was added multiple times
        message -- explanation of the error
    """

    def __init__(self, nonunique, message=" entered more than once. Please only enter command once"):
        self.nonunique = nonunique
        self.message = message
        print(self.nonunique + self.message)
        super().__init__()

class ArgumentError(HelpError):
    """
    Exception raised for errors in the command line arguments.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        print(self.message)
        super().__init__()

class LogReadError(HelpError):
    """
    Exception raised for errors in the reading log.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message
        print(self.message)
        super().__init__()


import random
import re

from .constants import ipv4_regex, ipv6_regex

def lighter(line):
    # https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters

    # make style bold
    style = 1

    # just make text black
    textcolor = 30


    bg = random.randint(40,47)
    if bg == 40:
        #background is black so change textcolor to white
        textcolor = 97

    # add bright colors 50% chance
    if random.randint(0, 1):
        bg += 60

    format = ';'.join([str(style), str(textcolor), str(bg)])

    return '\x1b[%sm%s\x1b[0m' % (format, line)

def highlighter_ipv4(results):
    pattern = re.compile(ipv4_regex)
    ipv4s = pattern.findall(results)
    for ip in set(ipv4s):
        results = results.replace(ip, lighter(ip))
    print(results)

def highlighter_ipv6(results):
    ipv6_address = re.compile(ipv6_regex)

    ipv6s = ipv6_address.findall(results)
    for ip in set(ipv6s):
        results = results.replace(ip, lighter(ip))
    print(results)

def highlight_both(results):
    ipv6_address = re.compile(ipv6_regex)

    ipv6s = ipv6_address.findall(results)
    for ip in set(ipv6s):
        results = results.replace(ip, lighter(ip))

    pattern = re.compile(ipv4_regex)
    ipv4s = pattern.findall(results)
    for ip in set(ipv4s):
        results = results.replace(ip, lighter(ip))
    print(results)


import mmap
import re
import ipaddress

from .highlight import highlighter_ipv4, highlighter_ipv6, highlight_both

class Parser():

    def __init__(self, file_obj, arguments):

        self.arguments = arguments

        # will this creat a copy, need to make sure it doesn't
        self.file_obj = file_obj
        self.num_lines = -1

        if self.arguments.mmap:
            self.engine = self.mmap_first
        else:
            self.engine = self.memory_first

        self.output = print

        # simple returning first or last lines
        self.simple = 1

        # first number of lines given
        if self.arguments.first:
            self.num_lines = self.arguments.first

        # last number of lines given
        if self.arguments.last:
            self.num_lines = self.arguments.last
            if self.arguments.mmap:
                self.engine = self.mmap_last
            else:
                self.engine = self.memory_last

        # parser is no longer simple, more parsing required
        if self.arguments.ipv6 or self.arguments.ipv4 or self.arguments.timestamps:
            self.simple = 0

            # set highlight
            if self.arguments.ipv4: self.output = highlighter_ipv4
            if self.arguments.ipv6: self.output = highlighter_ipv6
            if self.arguments.ipv6 and self.arguments.ipv4: self.output = highlight_both

    def followup(self, line):
        # search for timestamps
        if self.arguments.timestamps:
            timestamp_found = self.timestamps(line)
            if timestamp_found:
                # check to search for ip addresses
                if self.arguments.ipv4:
                    ipv4_found = self.ipv4(line)
                    if ipv4_found == 0 and self.arguments.ipv6:
                        return self.ipv6(line)
                    return ipv4_found
                # check but only for ipv6 addressses
                if self.arguments.ipv6:
                    return self.ipv6(line)
            # no timestamp found
            return timestamp_found

        # don't need to search for timestamps, only ip addresses
        if self.arguments.ipv4:
            ipv4_found = self.ipv4(line)
            if ipv4_found == 0 and self.arguments.ipv6:
                return self.ipv6(line)
            return ipv4_found

        if self.arguments.ipv6:
            return self.ipv6(line)

        # no follow up arguments
        return 0

    def mmap_first(self):
        try:
            with mmap.mmap(self.file_obj.fileno(), length=0, access=mmap.ACCESS_READ) as mmap_obj:
                file_length = len(mmap_obj)-1
                startpointer = 0
                endpointer = 0
                found_items=[]
                while self.num_lines != 0 and endpointer < file_length:
                    endpointer = mmap_obj.find(b'\n', startpointer, file_length) + 1
                    # end of file
                    if endpointer == 0:
                        endpointer = file_length+1

                    if self.simple:
                        self.num_lines -= 1
                    else:
                        # shift by one to be like first
                        success = self.followup(mmap_obj[startpointer:endpointer])
                        if success:
                            found_items.append([startpointer, endpointer])
                            self.num_lines -= 1


                    startpointer = endpointer

                if self.simple:
                    return mmap_obj[:endpointer].decode('utf-8')
                else:
                    result = ""
                    for item in found_items:
                        result += mmap_obj[item[0]:item[1]].decode('utf-8')
                    return result.rstrip('\n')

        except Exception as e:
            print(e)

    def mmap_last(self):
        try:
            with mmap.mmap(self.file_obj.fileno(), length=0, access=mmap.ACCESS_READ) as mmap_obj:
                file_length = len(mmap_obj) - 1
                startpointer = file_length
                endpointer = file_length+1 # to get last byte of end of file
                #have to track pointers to print logs in reverse order
                found_items = []
                # is it safe for startpointer > 0
                while self.num_lines != 0 and startpointer > 0:
                    startpointer = mmap_obj.rfind(b'\n', 0, endpointer)

                    if self.simple:
                        self.num_lines -= 1
                    else:
                        # shift by one to be like first
                        success = self.followup(mmap_obj[startpointer + 1:endpointer + 1])
                        if success:
                            found_items.insert(0, [startpointer + 1, endpointer + 1])
                            self.num_lines -= 1

                    endpointer = startpointer

                if self.simple:
                    return mmap_obj[startpointer+1:].decode('utf-8')
                else:
                    result = ""
                    for item in found_items:
                        result += mmap_obj[item[0]:item[1]].decode('utf-8')

                    return result.rstrip('\n')

        except Exception as e:
            print(e)

    def memory_first(self):
        try:
            file_length = len(self.file_obj) - 1
            startpointer = 0
            endpointer = 0
            found_items = []
            while self.num_lines != 0 and endpointer < file_length:
                endpointer = self.file_obj.find(b'\n', startpointer, file_length) + 1

                # end of file
                if endpointer == 0:
                    endpointer = file_length + 1

                if self.simple:
                    self.num_lines -= 1
                else:
                    # shift by one to be like first
                    success = self.followup(self.file_obj[startpointer:endpointer])
                    if success:
                        found_items.append([startpointer, endpointer])
                        self.num_lines -= 1

                startpointer = endpointer

            if self.simple:
                return self.file_obj[:endpointer].decode('utf-8')
            else:
                result = ""
                for item in found_items:
                    result += self.file_obj[item[0]:item[1]].decode('utf-8')
                return result.rstrip('\n')

        except Exception as e:
            print(e)

    def memory_last(self):
        try:
            file_length = len(self.file_obj) - 1
            startpointer = file_length
            endpointer = file_length
            # have to track pointers to print logs in reverse order
            found_items = []
            # is it safe for startpointer > 0
            while self.num_lines != 0 and startpointer > 0:
                startpointer = self.file_obj.rfind(b'\n', 0, endpointer)

                if self.simple:
                    self.num_lines -= 1
                else:
                    # shift by one to be like first
                    success = self.followup(self.file_obj[startpointer + 1:endpointer + 1])
                    if success:
                        found_items.insert(0, [startpointer + 1, endpointer + 1])
                        self.num_lines -= 1

                endpointer = startpointer

            if self.simple:
                return self.file_obj[startpointer + 1:].decode('utf-8')
            else:
                result = ""
                for item in found_items:
                    result += self.file_obj[item[0]:item[1]].decode('utf-8')
                return result.rstrip('\n')

        except Exception as e:
            print(e)

    def timestamps(self, line):

        if line:
            pattern = re.compile(rb'(?:[01]\d|2[0123]):(?:[012345]\d):(?:[012345]\d)')

            if len(pattern.findall(line)):
                return line

        # timestamp not found
        return 0

    def ipv4(self, line):

        if line:

            pattern = re.compile(rb'(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})')
            ips = pattern.findall(line)

            if len(ips):
                for ip in ips:
                    try:
                        ipaddress.IPv4Address(ip.decode('utf-8'))
                        # found ipv4
                        return line
                    except Exception as e:
                        pass

        # ipv4 not found
        return 0

    def ipv6(self, line):

        if line:
            pattern = rb'^.*((([0-9A-Fa-f]{1,4}:){7}([0-9A-Fa-f]{1,4}|:))|(([0-9A-Fa-f]{1,4}:){6}(:[0-9A-Fa-f]{1,4}|((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9A-Fa-f]{1,4}:){5}(((:[0-9A-Fa-f]{1,4}){1,2})|:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3})|:))|(([0-9A-Fa-f]{1,4}:){4}(((:[0-9A-Fa-f]{1,4}){1,3})|((:[0-9A-Fa-f]{1,4})?:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){3}(((:[0-9A-Fa-f]{1,4}){1,4})|((:[0-9A-Fa-f]{1,4}){0,2}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){2}(((:[0-9A-Fa-f]{1,4}){1,5})|((:[0-9A-Fa-f]{1,4}){0,3}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(([0-9A-Fa-f]{1,4}:){1}(((:[0-9A-Fa-f]{1,4}){1,6})|((:[0-9A-Fa-f]{1,4}){0,4}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:))|(:(((:[0-9A-Fa-f]{1,4}){1,7})|((:[0-9A-Fa-f]{1,4}){0,5}:((25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)(\.(25[0-5]|2[0-4]\d|1\d\d|[1-9]?\d)){3}))|:)))(%.+)?'

            if re.match(pattern, line):
                # found ipv6
                return line

        # ipv6 not found
        return 0

    def run(self):
        self.output(self.engine())
